from mlcore.datahelper.dynamodb.dynamodb_data_helper import DynamoDBDataHelper
from mlcore.datahelper.dynamodb.mllearningdb import create as Creator
from mlcore.datahelper.dynamodb.mllearningdb import update as Updator
from datetime import datetime
import json

from boto3.dynamodb.conditions import Key

class DataHelper(DynamoDBDataHelper):

    def __init__(
        self, 
        logger=None,
        config_file='go_servers.ini',
        host_name='ML_Learning_DEV'):
        super().__init__( 
            logger=logger,
            config_file=config_file,
            host_name=host_name)

    def connect(self):
        super().connect()
    
    def create_vm(self):
        table = self.create_table(Creator.vm)

        return table

    def create_battey_swap_log(self):
        table = self.create_table(Creator.battey_swap_log)

        return table

    def create_vm_status(self):
        table = self.create_table(Creator.vm_status)

        return table

    def update_battey_swap_log(self):
        table = self.update_table(Updator.battey_swap_log)

        return table

    def delete_vm(self):
        table = self.delete_table(Creator.vm['TableName'])

        return table

    def delete_battey_swap_log(self):
        table = self.delete_table(Creator.battey_swap_log['TableName'])

        return table

    def delete_vm_status(self):
        table = self.delete_table(Creator.vm_status)

        return table

    def query_battery_swap_log_by_date(self, dt_str):
        table = self.get_table(Creator.battey_swap_log['TableName'])
        filtering_exp = Key('ExchangeDate').eq(dt_str)
        response = table.query(KeyConditionExpression=filtering_exp)
        result = response['Items']
        if self.logger != None:
            self.logger.info('result.count = %s' % len(result))
            if len(result) > 0:
                self.logger.info('result[0] = %s' % result[0])

    def insert_vm(self, item_dict_list):
        item_list = []
        for item_dict in item_dict_list:
            dt = int(datetime.utcnow().timestamp())
            item = {
                'Id': item_dict['vm_guid'], 
                'Info': json.dumps(item_dict),
                'UpdateTime': dt}
            item_list.append(item)
            
        self.iterate_put_items(
            Creator.vm['TableName'],
            item_list)

    def insert_battey_swap_log(self, item_dict_list):
        if (len(item_dict_list))==0:
            self.logger.warn('Empty input of battey swap log')
            return

        item_list = []
        for item_dict in item_dict_list:
            [date_str, time_str] = self.timestamp_to_dt_str(item_dict['exchange_time'])
            item = {
                'ExchangeDate': date_str,
                'ExchangeTime#Id': '%s#%s' % (time_str, item_dict['_id']),
                'SwapLog': json.dumps(item_dict)}
            item_list.append(item)
        
        self.iterate_put_items(
            Creator.battey_swap_log['TableName'],
            item_list)

    def insert_vm_status(self, item_dict_list):
        item_list = []
        for item_dict in item_dict_list:
            item = {
                'Id': item_dict['vm_id'], 
                'Status': json.dumps(item_dict),
                'SnapTime': item_dict['snap_time']}
            item_list.append(item)
        
        self.iterate_put_items(
            Creator.vm_status['TableName'],
            item_list)
            
    def timestamp_to_dt_str(self, timestamp):
        '''
        Return a date_string and time_string by given timestamp
        '''
        dt_str = datetime.utcfromtimestamp(timestamp).strftime('%Y-%m-%d %H:%M:%S')
        return [dt_str[:10], dt_str[-8:]]




from mlcore.datahelper.api.keycloak_sdk import KeycloakSDK
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
import json
import base64
import sys
import logging

logging.basicConfig(stream=sys.stdout,
                    level=logging.INFO,
                    format='%(asctime)-s %(levelname)-7s %(name)-s.%(funcName)s(): %(message)s',
                    datefmt='%Y-%m-%dT%H:%M:%S')
                    
class APIHelper():

    def __init__(self, config):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.config = config['mlcore']['datahelper']['api']['mongodb']

        self.keycloakSDK = KeycloakSDK(self.config['keycloak'])
        self.keycloak = self.config['keycloak']
        self.access_token = None

    def get_access_token(self):
        return self.keycloakSDK.get_access_token()

    


import sys
import logging

logging.basicConfig(stream=sys.stdout,
                    level=logging.INFO,
                    format='%(asctime)-s %(levelname)-7s %(name)-s.%(funcName)s(): %(message)s',
                    datefmt='%Y-%m-%dT%H:%M:%S')

class Creator():
    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    #######################################################################################
    #
    #  Hive tables with partition
    #
    #######################################################################################
    def create_battery_swap_log(self, database, s3bucket='gogoro-ml-dev-model'):
        sql = '''
CREATE EXTERNAL TABLE IF NOT EXISTS %s.%s (
  `data` array< 
    struct< 
      `_id`: string,
      `create_time`: bigint,
      `exchange_time`: bigint,
      `swap_id`: bigint,
      `scooter_guid`: string,
      `vm_guid`: string,
      `rating`: double,
      `base_price`: double,
      `pps_price`: double,
      `rating_rule_id`: string,
      `eject_new_candidate`: boolean,
      `swap_in`:array<
        struct<
          `battery_guid`: string,
          `slot_id`: int,
          `slot_serial_id`: string,
          `healthy`: int,
          `cell_type`: int,
          `case_type`: int,
          `temp_cell`: int,
          `temp_mos`: int,
          `is_locked`: int,
          `fcc`: int,
          `consumption`: double,
          `charging_rate`: double,
          `max_charging_rate`: double,
          `chargeable`: boolean,
          `dischargeable`: boolean,
          `payload_raw`: string,
          `payload`: struct<
            `soc`: int,
            `capacity_level`: int,
            `initial_capacity`: int,
            `current_capacity`: int,
            `lifetime_charged_capacity`: int,
            `lifetime_discharged_capacity`: int,
            `fcc`: int,
            `max_cell_temp`: int,
            `min_cell_temp`: int,
            `recycle_count`: int
          >,
          `sgen_payload_version`: int,
          `hw_info_version`: int,
          `hw_info`: string,
          `ups`: int
      >>,
      `swap_out`:array<
        struct<
          `battery_guid`: string,
          `slot_id`: int,
          `slot_serial_id`: string,
          `healthy`: int,
          `cell_type`: int,
          `case_type`: int,
          `temp_cell`: int,
          `temp_mos`: int,
          `is_locked`: int,
          `fcc`: int,
          `consumption`: double,
          `charging_rate`: double,
          `max_charging_rate`: double,
          `chargeable`: boolean,
          `dischargeable`: boolean,
          `payload_raw`: string,
          `payload`: struct<
            `soc`: int,
            `capacity_level`: int,
            `initial_capacity`: int,
            `current_capacity`: int,
            `lifetime_charged_capacity`: int,
            `lifetime_discharged_capacity`: int,
            `fcc`: int,
            `max_cell_temp`: int,
            `min_cell_temp`: int,
            `recycle_count`: int
          >,
          `sgen_payload_version`: int,
          `hw_info_version`: int,
          `hw_info`: string,
          `ups`: int
      >>,
      scooter_data: struct<
        `swap_span`: int,
        `version`: string,
        `total_consumption`: double,
        `riding_distance`: double,
        `riding_history_raw`: string,
        `riding_history`: struct<
          `max_speed`: int,
          `avg_speed`: double,
          `idle_time`: int,
          `riding_time`: int,
          `total_consumption`: double
        >,
        `total_distance`: double
      >,
      trip_consumption: double
  >> 
) 
PARTITIONED BY (`dt` string)
ROW FORMAT SERDE 'org.openx.data.jsonserde.JsonSerDe'
WITH SERDEPROPERTIES ('serialization.format' = '1' ) 
LOCATION 's3://%s/ml/etl-workspace/battery-swap-log/' 
TBLPROPERTIES ('has_encrypted_data'='false');
        ''' % (database, 'battery_swap_log', s3bucket)

        return sql

    def create_vm_status(self, database, s3bucket='gogoro-ml-dev-model'):
        sql = '''
CREATE EXTERNAL TABLE IF NOT EXISTS %s.%s (
  `data` array< 
    struct< 
      `_id`: string,
      `create_time`: bigint,
      `snap_time`: bigint,
      `vm_guid`: string,
      `state`: int,
      `temperature`: int,
      `rack_temperature`: array<int>,
      `battery_single_count`: int,
      `battery_pair_count`: int,
      `power_meter`: double,
      `power_state`: int
  >> 
) 
PARTITIONED BY (`dt` string)
ROW FORMAT SERDE 'org.openx.data.jsonserde.JsonSerDe'
WITH SERDEPROPERTIES ('serialization.format' = '1' ) 
LOCATION 's3://%s/ml/etl-workspace/vm-status/' 
TBLPROPERTIES ('has_encrypted_data'='false');
        ''' % (database, 'vm_status', s3bucket)

        return sql

    def create_vm_battery_count(self, database, s3bucket='gogoro-ml-dev-model'):
        sql = '''
CREATE EXTERNAL TABLE IF NOT EXISTS %s.%s (
  `data` array< 
    struct< 
      `time_str`: string,
      `battery_counts`: array<
        struct<
          `vm_guid`: string,
          `battery_single_count`: int,
          `battery_pair_count`: int
      >>
  >> 
) 
PARTITIONED BY (`dt` string)
ROW FORMAT SERDE 'org.openx.data.jsonserde.JsonSerDe'
WITH SERDEPROPERTIES ('serialization.format' = '1' ) 
LOCATION 's3://%s/ml/etl-workspace/vm-battery-count/' 
TBLPROPERTIES ('has_encrypted_data'='false');
        ''' % (database, 'vm_battery_count', s3bucket)

        return sql

    def create_vm_exchange_count(self, database, s3bucket='gogoro-ml-dev-model'):
        sql = '''
CREATE EXTERNAL TABLE IF NOT EXISTS %s.%s (
  `data` array< 
    struct< 
      `vm_guid`: string,
      `swap_count`: int,
      `swap_battery_count`: int
  >> 
) 
PARTITIONED BY (`dt` string)
ROW FORMAT SERDE 'org.openx.data.jsonserde.JsonSerDe'
WITH SERDEPROPERTIES ('serialization.format' = '1' ) 
LOCATION 's3://%s/ml/etl-workspace/vm-exchange-count/' 
TBLPROPERTIES ('has_encrypted_data'='false');
        ''' % (database, 'vm_exchange_count', s3bucket)

        return sql

    def create_demand_prediction(self, database, s3bucket='gogoro-ml-dev-model'):
        sql = '''
CREATE EXTERNAL TABLE IF NOT EXISTS %s.%s (
  `vm_id` string,
  `prediction` double 
) PARTITIONED BY (
  dt string 
)
ROW FORMAT SERDE 'org.apache.hadoop.hive.serde2.lazy.LazySimpleSerDe'
WITH SERDEPROPERTIES (
  'serialization.format' = ',',
  'field.delim' = ','
) LOCATION 's3://%s/ml/job-workspace/demand-prediction/'
TBLPROPERTIES ('has_encrypted_data'='false');
        ''' % (database, 'demand_prediction', s3bucket)

    #######################################################################################
    #
    #  Hive tables without partition
    #
    #######################################################################################
#     def create_vm_list(self, database, s3bucket='gogoro-ml-dev-model'):
#         sql = '''
# CREATE EXTERNAL TABLE IF NOT EXISTS %s.%s (
#   `data` array< 
#     struct< 
#       `vm_id`: string,
#       `vm_guid`: string,
#       `vm_rid`: string,
#       `name`: struct <
#         `en-US`: string,
#         `zh-TW`: string
#       >,
#       `name_code`: int,
#       `latitude`: double,
#       `longitude`: double,
#       `address`: struct<
#         `en-US`: string,
#         `zh-TW`: string
#       >,
#       `district`: struct<
#         `en-US`: string,
#         `zh-TW`: string
#       >,
#       `city`: struct<
#         `en-US`: string,
#         `zh-TW`: string
#       >,
#       `zip_code`: int,
#       `country_code`: string,
#       `status`: int,
#       `battery_slot_count`: int,
#       `mainboard_guid`: string,
#       `charger_count`: int,
#       `power_supply_capacity`: int,
#       `available_time`: string,
#       `cell_config`: struct<
#         `CellConfigData`: array<
#           struct<
#             `CellType`: int,
#             `MaxChargeSpeed`: int,
#             `MaxReleaseTemperature`: int,
#             `SpeedLimitProfileType`: int
#         >>
#       >,
#       `update_time`: bigint,
#       `create_time`: bigint,
#       `battery_count`: int,
#       `battery_pair_count`: int
#   >> 
# ) 
# ROW FORMAT SERDE 'org.openx.data.jsonserde.JsonSerDe'
# WITH SERDEPROPERTIES ('serialization.format' = '1' ) 
# LOCATION 's3://%s/ml/rawdata-workspace/vm-list/latest/' 
# TBLPROPERTIES ('has_encrypted_data'='false');
#         ''' % (database, 'vm-list', s3bucket)

#         return sql

#     def create_vm_nearby(self, database, s3bucket='gogoro-ml-dev-model'):
#         sql = '''
# CREATE EXTERNAL TABLE IF NOT EXISTS %s.%s (
#   `data` array< 
#     struct< 
#       `vm_id`: string,
#       `nearby_vms`: array<
#           array<
#             `vm_id`: string,
#             `distance`: string,
#             `latitude`: double,
#             `longitude`: double,
#           >
#       >
# ) 
# ROW FORMAT SERDE 'org.openx.data.jsonserde.JsonSerDe'
# WITH SERDEPROPERTIES ('serialization.format' = '1' ) 
# LOCATION 's3://%s/ml/rawdata-workspace/vm-nearby/latest/' 
# TBLPROPERTIES ('has_encrypted_data'='false');
#         ''' % (database, 'vm-nearby', s3bucket)


from mlcore.datahelper.mongodb.mongodb_data_helper import MongoDBDataHelper
from mlcore.datahelper.mongodb.dataplatform import find as Finder
import sys
import logging

logging.basicConfig(stream=sys.stdout,
                    level=logging.INFO,
                    format='%(asctime)-s %(levelname)-7s %(name)-s.%(funcName)s(): %(message)s',
                    datefmt='%Y-%m-%dT%H:%M:%S')

class DataHelper(MongoDBDataHelper):

    def __init__(self, config):

        self.config = config['mlcore']['datahelper']['mongodb']['dataplatform']
        super().__init__(self.config)
        self.logger = logging.getLogger(self.__class__.__name__)

    def connect(self):
        super().connect()

    #def get_battery_swap_log_b(self, start_ts, end_ts=0):
    #    return super().find(
    #        Finder.CollectionNames().battery_swap_log_b, 
    #        Finder.get_battery_swap_log(start_ts, end_ts))

    def get_battery_swap_log(self, start_ts, end_ts=0):
        return super().find(
            Finder.CollectionNames().battery_swap_log, 
            Finder.get_battery_swap_log(start_ts, end_ts))

    #def get_vm_status_b(self, start_ts, end_ts=0):
    #    return super().find(
    #        Finder.CollectionNames().vm_status_b, 
    #        Finder.get_vm_status(start_ts, end_ts))

    def get_vm_status(self, start_ts, end_ts=0):
        return super().find(
            Finder.CollectionNames().vm_status, 
            Finder.get_vm_status(start_ts, end_ts))



update_ts_interval = 24 * 60 * 60       # 1 day


class CollectionNames():

    def __init__(self):
        self.battery_swap_log_b = 'battery_swap_log_b'
        self.battery_swap_log = 'battery_swap_log'
        self.vm_status_b = 'vm_status_b'
        self.vm_status = 'vm_status'


def get_battery_swap_log(start_ts, end_ts=0):
    global update_ts_interval
    if end_ts==0:
        return {
            'exchange_time': {
                '$gte': start_ts,
                '$lt': start_ts + update_ts_interval
            }
        }
    else:
        return {
            'exchange_time': {
                '$gte': start_ts,
                '$lt': end_ts
            }
        }

def get_vm_status(start_ts, end_ts=0):
    global update_ts_interval
    if end_ts==0:
        return {
            'snap_time': {
                '$gte': start_ts,
                '$lt': start_ts + update_ts_interval
            }
        }
    else:
        return {
            'snap_time': {
                '$gte': start_ts,
                '$lt': end_ts
            }
        }



import pytest
from mlcore.datahelper.s3.mllearningdb.s3_path_helper import RawDataPathHelper


class TestRawDataPathHelper:
    # def test_get_rawdata_key(self):
    #     assert RawDataPathHelper().get_rawdata_key('vm_list') == 'vm-list'

    def test_get_dirs_by_name(self):
        path = RawDataPathHelper().get_dirs_by_name('vm_list')
        assert path is not None
        assert 's3folder' in path
        assert 'local_dir' in path
        assert 's3prefix' in path
        assert 'local_prefix' in path

        s3_root_dir = RawDataPathHelper().s3_root_dir
        assert s3_root_dir in path['s3folder'] 
        assert s3_root_dir in path['s3prefix'] 
        local_root_dir = RawDataPathHelper().local_root_dir
        assert local_root_dir in path['local_dir'] 
        assert local_root_dir in path['local_prefix'] 


import boto3
import botocore
import os
import shutil
import random

import sys
import logging

logging.basicConfig(stream=sys.stdout,
                    level=logging.INFO,
                    format='%(asctime)-s %(levelname)-7s %(name)-s.%(funcName)s(): %(message)s',
                    datefmt='%Y-%m-%dT%H:%M:%S')

class S3DataHelper():

    def __init__(self, config):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.config = config
        self.connect()

    def connect(self):
        ###################################################
        '''
        return s3 resource
        '''
        def get_s3_resource(config):    
            if 'role-arn' in config and config['role-arn']!='':
                if 'aws-access-key-id' in config and 'aws-secret-access-key' in config:
                    sts_client = boto3.client(
                        'sts',
                        aws_access_key_id=config['aws-access-key-id'],
                        aws_secret_access_key=config['aws-secret-access-key'],
                        region_name=config['region'])
                else:
                    sts_client = boto3.client('sts')

                session_name = '%s-%s'%(config['role-session-name'], str(random.randint(1,100)))
                #self.logger.info('session_name = %s' % session_name)
                assumed_role_object = sts_client.assume_role(
				    RoleArn=config['role-arn'],
				    RoleSessionName=session_name)
                credentials = assumed_role_object['Credentials']
            
                s3 = boto3.resource(
				    's3',
				    aws_access_key_id=credentials['AccessKeyId'],
				    aws_secret_access_key=credentials['SecretAccessKey'],
				    aws_session_token=credentials['SessionToken'],
                    region_name=config['region'])
            else:
                if 'aws-access-key-id' in config and 'aws-secret-access-key' in config:
                    s3 = boto3.resource(
                        's3',
                        aws_access_key_id=config['aws-access-key-id'],
                        aws_secret_access_key=config['aws-secret-access-key'],
                        region_name=config['region'])
                else:
                    s3 = boto3.resource(
                        's3',
                        region_name=config['region'])

            if 's3baseprefix' in config:
                return {
                    's3': s3,
                    's3bucket': config['s3bucket'],
                    's3baseprefix': config['s3baseprefix']
                }
            else:
                return {
                    's3': s3,
                    's3bucket': config['s3bucket'],
                    's3baseprefix': 'ml/'
                }
        ###################################################
        
        # s3 read workspace
        if 'read' in self.config:
            self.s3read = get_s3_resource(self.config['read'])
        else:
            self.s3read = None
        # s3 write workspace
        if 'write' in self.config:
            self.s3write = get_s3_resource(self.config['write'])
        else:
            self.s3write = self.s3read

    def is_object_exist(self, key, is_write=False):
        '''
        check if an object in exist (True) or not (False)
        '''
        self.connect()
        s3 = self.s3read
        if is_write==True:
            s3 = self.s3write

        response = s3['s3'].meta.client.list_objects_v2(
            Bucket=s3['s3bucket'],
            Prefix=key,
        )
        for obj in response.get('Contents', []):
            if obj['Key'] == key:
                return True
        return False

    def list_object(self, prefix, is_write=False):
        self.connect()
        s3 = self.s3read
        if is_write==True:
            s3 = self.s3write

        response = s3['s3'].meta.client.list_objects_v2(
            Bucket=s3['s3bucket'],
            Prefix=prefix)
        if 'Contents' in response:
            return response['Contents']
        else:
            self.logger.warn('No contents in s3(Bucket=%s, prefix=%s)' % (s3['s3bucket'], prefix))
            return []

    def get_keys(self, prefix, is_write=False):
        """Get a list of all keys in an S3 bucket."""

        self.connect()
        s3 = self.s3read
        if is_write==True:
            s3 = self.s3write

        keys = []

        kwargs = {'Bucket': s3['s3bucket'], 'Prefix': prefix}
        while True:
            resp = s3['s3'].meta.client.list_objects_v2(**kwargs)
            if 'Contents' not in resp:
                break
            for obj in resp['Contents']:
                keys.append(obj['Key'])

            try:
                kwargs['ContinuationToken'] = resp['NextContinuationToken']
            except KeyError:
                break
        
        if len(keys)==0:
            self.logger.warn('No contents in s3(Bucket=%s, prefix=%s)' % (s3['s3bucket'], prefix))

        return keys

    def get_folder_prefixes(self, prefix, is_write=False):
        self.connect()
        s3 = self.s3read
        if is_write==True:
            s3 = self.s3write

        keys = []
        kwargs = {'Bucket': s3['s3bucket'], 'Prefix': prefix, 'Delimiter': '/'}

        while True:
            resp = s3['s3'].meta.client.list_objects_v2(**kwargs)
            if 'CommonPrefixes' not in resp:
                break

            for obj in resp['CommonPrefixes']:
                keys.append(obj['Prefix'])

            try:
                kwargs['ContinuationToken'] = resp['NextContinuationToken']
            except KeyError:
                break
        
        if len(keys)==0:
            self.logger.warn('No folders in s3(Bucket=%s, prefix=%s)' % (s3['s3bucket'], prefix))

        return keys

    def delete_files(self, list_s3path, max_keys=1000):
        '''
        This operation enables you to delete multiple objects from a bucket using a single HTTP request. 
        You may specify up to 1000 keys.
        '''
        self.connect()

        self.logger.info('delete files in s3 (bucket=%s, key=%s)' % (self.s3write['s3bucket'], list_s3path))

        n = len(list_s3path)
        cur = 0
        while cur < n:
            objs = [ {'Key':s3path} for s3path in list_s3path[cur:cur+max_keys]]
            self.s3write['s3'].meta.client.delete_objects(
                Bucket=self.s3write['s3bucket'],
                Delete={
                    'Objects': objs
                }
            )
            cur += max_keys
        # self.s3write['s3'].meta.client.delete_object(
        #     Bucket=self.s3write['s3bucket'],
        #     Key=list_s3path
        # )

    def delete_folder(self, s3folder):
        '''
        delete a s3 folder
        '''
        if s3folder[-1] != '/':
            self.logger.warn('not a valid s3folder %s' % s3folder)
            return

        self.connect()

        self.logger.info('delete folder in s3 (bucket=%s, key=%s)' % (self.s3write['s3bucket'], s3folder))

        objs = self.list_object(s3folder, is_write=True)
        if len(objs)==0:
            return

        list_s3path = []
        for obj in objs:
            list_s3path.append(obj['Key'])
            #self.delete_file(s3path)
        self.delete_files(list_s3path)

        

    def upload_file(self, s3path, local_file):
        self.connect()
        self.logger.info('put %s to s3 (bucket=%s, key=%s)' % (local_file, self.s3write['s3bucket'], s3path))

        self.s3write['s3'].meta.client.upload_file(
            local_file,
            self.s3write['s3bucket'],
            s3path)

    def upload_folder(self, s3folder, local_dir):
        '''
        upload a local folder to S3 (with success file)
        '''

        if s3folder[-1] != '/':
            self.logger.warn('not a valid s3folder %s' % s3folder)
            return
        if local_dir[-1] != '/':
            self.logger.warn('not a valid local_dir %s' % local_dir)
            return

        self.connect()
        self.logger.info('put %s to s3 (bucket=%s, key=%s)' % (local_dir, self.s3write['s3bucket'], s3folder))

        files = os.listdir(local_dir)
        err_cnt = 0
        for file in files:
            local_file = '%s%s' % (local_dir, file)
            s3_path = '%s%s' % (s3folder, file)
            try: 
                self.upload_file(s3_path, local_file)
            except Exception as e:
                self.logger.warning("failed to upload file %s to s3://%s/%s" % (path, self.s3write['s3bucket'], s3_path))
                self.logger.warning(e)
                err_cnt += 1

        if err_cnt==0:
            success_file_path = '%s%s' % (s3folder, '_SUCCESS')
            self.logger.info("put _SUCCESS to s3 (bucket=%s, key=%s)" % (self.s3write['s3bucket'], success_file_path))
            self.s3write['s3'].meta.client.put_object(
                    Bucket=self.s3write['s3bucket'],
                    Key=success_file_path)

    def download_file(self, s3path, local_file):
        self.connect()
        self.logger.info('get %s from s3 (bucket=%s, key=%s)' % (local_file, self.s3read['s3bucket'], s3path))
        try:
            self.mkdir("/".join(local_file.split('/')[:-1]) + "/")
            self.s3read['s3'].Bucket(self.s3read['s3bucket']).download_file(s3path, local_file)
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "404":
                self.logger.warn("The object in s3 (bucket=%s, key=%s) does not exist." % (self.s3read['s3bucket'], s3path))
            else:
                raise

    def download_folder(self, s3folder, local_dir):
        '''
        dowload a s3 folder to a local folder 
        '''

        if s3folder[-1] != '/':
            self.logger.warn('not a valid s3folder %s' % s3folder)
            return
        if local_dir[-1] != '/':
            self.logger.warn('not a valid local_dir %s' % local_dir)
            return

        self.connect()
        self.logger.info('get %s from s3 (bucket=%s, key=%s)' % (local_dir, self.s3read['s3bucket'], s3folder))

        objs = self.list_object(s3folder)
        if len(objs)==0:
            return

        self.mkdir(local_dir)
        for obj in objs:
            s3path = obj['Key']
            local_file = '%s%s'% (local_dir, s3path[len(s3folder):])
            #self.logger.info('local_dir=%s, s3path=%s, s3path[len(s3folder):]=%s' % (local_dir, s3path, s3path[len(s3folder):]))
            self.download_file(s3path, local_file)

    def mkdir(self, root_dir):
        try :
            if not os.path.exists(root_dir):
                self.logger.info('create local dir: %s' % root_dir)
                os.makedirs(root_dir)
        except e:
            raise

    def rmdir(self, root_dir):
        '''
        Caution!
        It's dangerous like 'rm -rf'
        '''
        if root_dir[0]!='/':
            shutil.rmtree(root_dir)


    

    




import pytest
from mlcore.datahelper.s3.mllearningdb_data_helper import DataHelper
from config.get_config import get_config

class TestDataHelper:

    def test_get_latest_s3key(self):
        config = get_config('./config/config_dev.yml')
        data_helper = DataHelper(config)

        keys = [
            'test-workspace/2019-01-14T07:42:47/data.json',
            'test-workspace/2019-01-14T07:42:47/_SUCCESS',
            'test-workspace/2019-01-14T08:42:47/data.json',
            'test-workspace/2019-01-14T08:42:47/_SUCCESS',
            'test-workspace/2019-01-13T09:42:47/data.json',
            'test-workspace/2019-01-13T09:42:47/_SUCCESS',
            'test-workspace/2019-01-15T13:42:47/data.json'
        ]
        assert DataHelper(config).get_latest_s3key(keys) == 'test-workspace/2019-01-14T08:42:47/data.json'

    def test_get_latest_job_s3key(self):
        config = get_config('./config/config_dev.yml')
        data_helper = DataHelper(config)

        keys = [
            'test-workspace/2019-01-14T07:42:47/data.json',
            'test-workspace/2019-01-14T07:42:47/_SUCCESS',
            'test-workspace/2019-01-14T08:42:47/data.json',
            'test-workspace/2019-01-14T08:42:47/_SUCCESS',
            'test-workspace/2019-01-13T09:42:47/data.json',
            'test-workspace/2019-01-13T09:42:47/_SUCCESS',
            'test-workspace/2019-01-15T13:42:47/data.json'
        ]
        assert DataHelper(config).get_latest_job_s3key(keys) == 'test-workspace/2019-01-14T08:42:47/data.json'
    
    def test_get_latest_etl_s3key(self):
        config = get_config('./config/config_dev.yml')
        data_helper = DataHelper(config)

        keys = [
            'test-workspace/2019-01-14T07:42:47/data.json',
            'test-workspace/2019-01-14T07:42:47/_SUCCESS',
            'test-workspace/2019-01-14T08:42:47/data.json',
            'test-workspace/2019-01-14T08:42:47/_SUCCESS',
            'test-workspace/2019-01-13T09:42:47/data.json',
            'test-workspace/2019-01-13T09:42:47/_SUCCESS',
            'test-workspace/2019-01-15T13:42:47/data.json'
        ]
        assert DataHelper(config).get_latest_etl_s3key(keys) == 'test-workspace/2019-01-14T08:42:47/data.json'


    def test_get_dt_str_by_timestamp(self):
        config = get_config('./config/config_dev.yml')
        data_helper = DataHelper(config)


        dt_str = data_helper.get_dt_str_by_timestamp(1547451767)
        assert dt_str['dt'] == '2019-01-14T07:42:47'
        assert dt_str['dt_hour_str'] == '2019-01-14T07'
        assert dt_str['dt_minute_str'] == '2019-01-14T07:42'
        assert dt_str['time_str'] == '07:42:47'




import pyodbc
import configparser
from collections import OrderedDict

class MSSQLDataHelper():

    def __init__(
        self, 
        logger=None,
        config_file='go_servers.ini',
        host_name='GoServer_Production_ReadOnly'):
        self.logger = logger
        self.config_file = config_file
        self.host_name = host_name

    def connect(self):
        self.config = configparser.ConfigParser()
        self.config.read(self.config_file)
        conn_str = 'DRIVER={ODBC Driver 13 for SQL Server};SERVER=%s;DATABASE=%s;UID=%s;PWD=%s;PORT=%s' \
            % (self.config[self.host_name]['Server'], \
            self.config[self.host_name]['Database'], \
            self.config[self.host_name]['Username'], \
            self.config[self.host_name]['Password'], \
            self.config[self.host_name]['PORT'])
        self.conn = pyodbc.connect(conn_str)
        return self.conn

    def get_data(self, sql_str):
        self.logger_info('start get data from %s ... '% self.host_name)
        self.logger_info('\t%s' % sql_str)
        self.connect()
        cursor = self.conn.cursor()
        result = []
    
        try:
            cursor.execute(sql_str)
            columns = [desc[0] for desc in cursor.description]
            row = cursor.fetchone()
            while row is not None:
                row_dict = OrderedDict(zip(columns, row))
                result.append(dict(row_dict))
                row = cursor.fetchone()
        except Exception as e:
            self.logger_info(e)
            raise e
        finally:
            cursor.close()
            self.conn.close()

        self.logger_info('result.count = %s' % len(result))
        if len(result) > 0:
            self.logger_info('result[0] = %s' % result[0])
            #self.logger_info('result[0] = %s' % dict(result[0]))

        return result

    def get_vm_list(self):
        sql_str = """
SELECT Id, Latitude, Longitude, LocName
FROM dbo.Vm vm
LEFT JOIN dbo.VmExt vmExt ON vm.Id = vmExt.vmId
WHERE (STATE = 1 OR STATE = 0 OR STATE = 99 OR STATE = 100)
    AND vmExt.vmId IS NULL
        """
        
        res = self.get_data(sql_str)
        # filter by LocName
        for r in res:
            if '測試' in r['LocName'] or 'Test' in r['LocName'] or '建置' in r['LocName']:
                res.remove(r)

        self.logger_info('result.count = %s' % len(res))
        if len(res) > 0:
            self.logger_info('result[0] = %s' % res[0])

        return res

    def get_exchange_battery_diff(self, dt_start, dt_end):
        sql_str = """
SELECT exitem.ExId, MIN(exitem.Capacity) AS MinCapacity, MAX(exitem.Capacity) AS MaxCapacity
FROM dbo.BatteryExHistory ex
LEFT OUTER JOIN dbo.BatteryExItemHistory exitem ON ex.Id = exitem.ExId
WHERE ex.ExchangeTime >= '%s' 
    AND ex.ExchangeTime < '%s'
    AND VmId IS NOT NULL 
    AND ScooterId IS NOT NULL
    AND exitem.Type = 2
GROUP BY exitem.ExId
    """ % (dt_start, dt_end)

        res = self.get_data(sql_str)

        self.logger_info('result.count = %s' % len(res))
        if len(res) > 0:
            self.logger_info('result[0] = %s' % res[0])

        return res

    def get_scooter_miss_rate(self, dt_start, dt_end, r_min=0, r_max=85):
        sql_str = """
SELECT ex.ScooterId, COUNT(*) AS MissBatteries, COUNT(DISTINCT ExchangeTime) AS MissCount
FROM dbo.BatteryExHistory ex
LEFT OUTER JOIN dbo.BatteryExItemHistory exitem ON ex.Id = exitem.ExId
WHERE ex.ExchangeTime >= '%s' 
    AND ex.ExchangeTime < '%s'
    AND VmId IS NOT null 
    AND ScooterId IS NOT NULL
    AND exitem.Type = 2
    AND (exitem.Capacity >= %s AND exitem.Capacity <= %s)
GROUP BY ex.ScooterId
        """ % (dt_start, dt_end, r_min, r_max)

        res = self.get_data(sql_str)

        self.logger_info('result.count = %s' % len(res))
        if len(res) > 0:
            self.logger_info('result[0] = %s' % res[0])

        return res

    def get_vm_miss_rate(self, dt_start, dt_end):
        sql_str = """
SELECT ex.*, vm.Latitude, vm.Longitude
FROM (
    SELECT ex.VmId, COUNT(*) AS MissBatteries, COUNT(DISTINCT ExchangeTime) AS MissCount
    FROM dbo.BatteryExHistory ex
    LEFT OUTER JOIN dbo.BatteryExItemHistory exitem ON ex.Id = exitem.ExId
    WHERE ex.ExchangeTime >= '%s' 
        AND ex.ExchangeTime < '%s'
        AND VmId IS NOT null 
        AND ScooterId IS NOT NULL
        AND exitem.Type = 2
        AND exitem.Capacity < 85
GROUP BY ex.VmId
) AS ex
LEFT JOIN dbo.vm vm ON vm.id = ex.VmId;
        """ % (dt_start, dt_end)

        res = self.get_data(sql_str)

        self.logger_info('result.count = %s' % len(res))
        if len(res) > 0:
            self.logger_info('result[0] = %s' % res[0])

        return res

    def get_vm_total_exchange(self, dt_start, dt_end):
        sql_str = """
SELECT ex.*, vm.RId, vm.LocName, vm.Latitude, vm.Longitude, vm.Address
FROM (
    SELECT ex.VmId, COUNT(*) AS TotalBatteries, COUNT(DISTINCT ExchangeTime) AS TotalCount
    FROM dbo.BatteryExHistory ex
    LEFT OUTER JOIN dbo.BatteryExItemHistory exitem ON ex.Id = exitem.ExId
    WHERE ex.ExchangeTime >= '%s' 
        AND ex.ExchangeTime < '%s'
        AND VmId IS NOT null 
        AND ScooterId IS NOT NULL
        AND exitem.Type = 2
GROUP BY ex.VmId
) AS ex
LEFT JOIN dbo.vm vm ON vm.id = ex.VmId;
        """ % (dt_start, dt_end)

        res = self.get_data(sql_str)

        self.logger_info('result.count = %s' % len(res))
        if len(res) > 0:
            self.logger_info('result[0] = %s' % res[0])

        return res        

    def logger_info(self, msg):
        if self.logger != None:
            self.logger.info(msg)



from mlcore.datahelper.s3.s3_data_helper import S3DataHelper
from mlcore.datahelper.s3.mllearningdb.s3_path_helper import ETLPathHelper, JobPathHelper, RawDataPathHelper
from datetime import datetime
import json
import os

import sys
import logging

logging.basicConfig(stream=sys.stdout,
                    level=logging.INFO,
                    format='%(asctime)-s %(levelname)-7s %(name)-s.%(funcName)s(): %(message)s',
                    datefmt='%Y-%m-%dT%H:%M:%S')


class DataHelper(S3DataHelper):

    def __init__(self, config):

        self.config = config['mlcore']['datahelper']['s3']['mllearningdb']
        super().__init__(self.config)
        self.logger = logging.getLogger(self.__class__.__name__)

    def connect(self):
        super().connect()

    #####################################################
    #
    #   updated data
    #
    #####################################################
    
    def get_latest_s3folder(self, s3folders):
        '''
        get latest s3 folder with _SUCCESS file
        '''
        if len(s3folders)==0:
            return None

        s3folders.sort(key=lambda x:x, reverse=True)
        for i in range(0, len(s3folders)):
            # if (Not _SUCCESS file) and (have corresponding _SUCCESS file)
            s3folder = s3folders[i]
            s3keys = self.get_keys(s3folder)
            s3_success_key = '%s%s' % (s3folder, '_SUCCESS')
            #self.logger.info('s3keys = %s, s3_success_key=%s' % (s3keys, s3_success_key))
            if s3_success_key in s3keys:
                return s3folder

        return None

    def write_es_rates(self, data_dict):
        if len(data_dict)==0:
            self.logger.warn('Empty input of es rate')
            return

        # upload data to s3
        dt_now_str = self.get_dt_str_by_timestamp(int(datetime.now().timestamp()))['dt']
        path = RawDataPathHelper(s3baseprefix=self.s3write['s3baseprefix']).get_es_rates_dirs(update_time_str=dt_now_str)

        # write local file and upload
        self.mkdir(path['local_dir'])

        json_data = {}
        json_data['data'] = data_dict
        with open('%s%s.json' % (path['local_dir'], dt_now_str), 'w') as fp:
            json.dump(json_data, fp)
        # upload 
        self.upload_folder(path['s3folder'], path['local_dir'])
        # FIXME
        self.delete_folder(path['s3folder_latest'])
        self.upload_folder(path['s3folder_latest'], path['local_dir'])
        # remove local file
        self.rmdir(path['local_dir']) 


    def write_vm_list(self, data_dict):
        if len(data_dict)==0:
            self.logger.warn('Empty input of vm list')
            return

        # upload data to s3
        dt_now_str = self.get_dt_str_by_timestamp(int(datetime.now().timestamp()))['dt']
        path = RawDataPathHelper(s3baseprefix=self.s3write['s3baseprefix']).get_vm_list_dirs(update_time_str=dt_now_str)

        # write local file and upload
        self.mkdir(path['local_dir'])

        json_data = {}
        json_data['data'] = data_dict
        with open('%s%s.json' % (path['local_dir'], dt_now_str), 'w') as fp:
            json.dump(json_data, fp)
        # upload
        self.upload_folder(path['s3folder'], path['local_dir'])
        # FIXME
        self.delete_folder(path['s3folder_latest'])
        self.upload_folder(path['s3folder_latest'], path['local_dir'])
        # remove local file
        self.rmdir(path['local_dir'])  

    def write_vm_nearby(self, data_dict):
        if len(data_dict)==0:
            self.logger.warn('Empty input of vm nearby')
            return

        # upload data to s3
        dt_now_str = self.get_dt_str_by_timestamp(int(datetime.now().timestamp()))['dt']
        path = RawDataPathHelper(s3baseprefix=self.s3write['s3baseprefix']).get_vm_nearby_dirs(update_time_str=dt_now_str)

        # write local file and upload
        self.mkdir(path['local_dir'])

        json_data = {}
        json_data['data'] = data_dict
        with open('%s%s.json' % (path['local_dir'], dt_now_str), 'w') as fp:
            json.dump(json_data, fp)
        # upload
        self.upload_folder(path['s3folder'], path['local_dir'])
        # FIXME
        self.delete_folder(path['s3folder_latest'])
        self.upload_folder(path['s3folder_latest'], path['local_dir'])
        # remove local file
        self.rmdir(path['local_dir'])  

    def get_es_rates_s3folder(self):
        path = RawDataPathHelper(s3baseprefix=self.s3read['s3baseprefix']).get_es_rates_dirs()
        s3folder_latest = path['s3folder_latest']
        if self.is_object_exist('%s%s' % (s3folder_latest, '_SUCCESS')) == True:
            # return latest
            s3folder = s3folder_latest
        else:
            # no latest => search snapshot
            prefix = path['s3prefix']
            s3folders = self.get_folder_prefixes(prefix)
            s3folder = self.get_latest_s3folder(s3folders)

        if s3folder == None:
            self.logger.warn('No es_rates')
            return None
        else:
            self.logger.info('get es_rates s3folder = %s' % s3folder)
            return s3folder

    def get_vm_list_s3folder(self):
        path = RawDataPathHelper(s3baseprefix=self.s3read['s3baseprefix']).get_vm_list_dirs()
        s3folder_latest = path['s3folder_latest']
        if self.is_object_exist('%s%s' % (s3folder_latest, '_SUCCESS')) == True:
            # return latest
            s3folder = s3folder_latest
        else:
            # no latest => search snapshot
            prefix = path['s3prefix']
            s3folders = self.get_folder_prefixes(prefix)
            s3folder = self.get_latest_s3folder(s3folders)
        
        if s3folder == None:
            self.logger.warn('No vm_list')
            return None
        else:
            self.logger.info('get vm_list s3folder = %s' % s3folder)
            return s3folder

    def get_vm_nearby_s3folder(self):
        path = RawDataPathHelper(s3baseprefix=self.s3read['s3baseprefix']).get_vm_nearby_dirs()
        s3folder_latest = path['s3folder_latest']
        if self.is_object_exist('%s%s' % (s3folder_latest, '_SUCCESS')) == True:
            # return latest
            s3folder = s3folder_latest
        else:
            # no latest => search snapshot
            prefix = path['s3prefix']
            s3folders = self.get_folder_prefixes(prefix)
            s3folder = self.get_latest_s3folder(s3folders)

        if s3folder == None:
            self.logger.warn('No vm_nearby')
            return None
        else:
            self.logger.info('get vm_nearby s3folder = %s' % s3folder)
            return s3folder

    def get_es_rates(self):
        s3folder = self.get_es_rates_s3folder()

        if s3folder is None:
            self.logger.warn('No es_rates')
            return []
        
        path = RawDataPathHelper(s3baseprefix=self.s3read['s3baseprefix']).get_es_rates_dirs(
            update_time_str=s3folder.split('/')[-2])

        self.download_folder(path['s3folder'], path['local_dir'])

        # read local files
        files = os.listdir(path['local_dir'])
        data = None
        for file in files:
            if file[-5:] != '.json':
                continue
            local_path = '%s%s' % (path['local_dir'], file)
            with open(local_path) as fp:
                json_data = json.load(fp)
                data = json_data['data']

        # remove local files
        self.rmdir(path['local_dir'])

        self.logger.info('Read es_rates = %s' % data)
        return data

    def get_vm_list(self):
        s3folder = self.get_vm_list_s3folder()

        if s3folder is None:
            self.logger.warn('No vm_list')
            return []
        
        path = RawDataPathHelper(s3baseprefix=self.s3read['s3baseprefix']).get_vm_list_dirs(
            update_time_str=s3folder.split('/')[-2])

        self.download_folder(path['s3folder'], path['local_dir'])

        # read local files
        files = os.listdir(path['local_dir'])
        data = None
        for file in files:
            if file[-5:] != '.json':
                continue
            local_path = '%s%s' % (path['local_dir'], file)
            with open(local_path) as fp:
                json_data = json.load(fp)
                data = json_data['data']
        
        # remove local files
        self.rmdir(path['local_dir'])

        self.logger.info('Read %s logs from vm_list' % (len(data)))
        
        if len(data)==0:
            self.logger.warn('Empty vm_list.')
        else:
            self.logger.info('vm_list[0] = %s' % data[0])
        return data

    def get_vm_nearby(self):
        s3folder = self.get_vm_nearby_s3folder()

        if s3folder is None:
            self.logger.warn('No vm_nearby')
            return []
        
        path = RawDataPathHelper(s3baseprefix=self.s3read['s3baseprefix']).get_vm_nearby_dirs(
            update_time_str=s3folder.split('/')[-2])

        self.download_folder(path['s3folder'], path['local_dir'])

        # read local files
        files = os.listdir(path['local_dir'])
        data = None
        for file in files:
            if file[-5:] != '.json':
                continue
            local_path = '%s%s' % (path['local_dir'], file)
            with open(local_path) as fp:
                json_data = json.load(fp)
                data = json_data['data']

        # remove local files
        self.rmdir(path['local_dir'])

        self.logger.info('Read %s logs from vm_nearby' % (len(data)))
        if len(data)==0:
            self.logger.warn('Empty vm_nearby.')
        else:
            self.logger.info('vm_nearby[0] = %s' % data[0])
        return data

    #####################################################
    #
    #   jobs
    #
    #####################################################

    def filter_by_time(self, keys, filter_time_str):
        filter_keys = []
        for key in keys:
            time_str = key.split('/')[-2]
            if datetime.strptime(time_str, '%Y-%m-%dT%H:%M:%S') <= datetime.strptime(filter_time_str, '%Y-%m-%dT%H:%M:%S'):
                filter_keys.append(key)
        return filter_keys

    def get_latest_job_s3key(self, job_keys):
        if len(job_keys)==0:
            return None
        job_keys.sort(key=lambda x:x, reverse=True)
        for i in range(0, len(job_keys)):
            # if (Not _SUCCESS file) and (have corresponding _SUCCESS file)
            s3key = job_keys[i]
            s3_success_key = '%s%s' % ('/'.join(s3key.split('/')[:-1]), '/_SUCCESS')
            if s3key[-8:]!='_SUCCESS' and s3_success_key in job_keys:
                return s3key

        return None 

    def get_latest_job_s3folder(self, s3folders, update_time_str=''):
        '''
        get latest s3 folder with _SUCCESS file
        '''
        if len(s3folders)==0:
            return None

        def filter_folder_by_time(s3folders, filter_time_str):
            if filter_time_str=='':
                return s3folders

            filter_s3folders = []
            for s3folder in s3folders:
                time_str = s3folder.split('/')[-2]
                if datetime.strptime(time_str, '%Y-%m-%dT%H:%M:%S') <= datetime.strptime(filter_time_str, '%Y-%m-%dT%H:%M:%S'):
                    filter_s3folders.append(s3folder)
            return filter_s3folders
 
        s3folders = filter_folder_by_time(s3folders, update_time_str)

        return self.get_latest_s3folder(s3folders)

    def get_demand_prediction_result_s3folder(self, job_name, dt_hour_str, update_time_str=''):
        '''
        get latest (before 'update_time_str') demand prediction result in given 'dt_hour_str'
        '''
        job_path = JobPathHelper(s3baseprefix=self.s3read['s3baseprefix']).get_dirs_by_demand_prediction_name(job_name, dt_hour_str)
        job_prefix = job_path['s3prefix']
        s3folders = self.get_folder_prefixes(job_prefix)
        
        s3folder = self.get_latest_job_s3folder(s3folders, update_time_str=update_time_str)

        # return a folder 
        self.logger.info('get demand_prediction_result s3folder = %s' % s3folder)
        return s3folder

    def download_demand_prediction_result_by_job_name(
        self, 
        job_name, 
        dt_hour_str, 
        update_time_str=''):

        s3folder = self.get_demand_prediction_result_s3folder(
            job_name,
            dt_hour_str,
            update_time_str=update_time_str)

        if s3folder is None:
            if update_time_str=='':
                self.logger.warn('No demand_prediction_result in %s for %s' % (dt_hour_str, job_name))
            else:
                self.logger.warn('No demand_prediction_result in %s for %s (before %s)' % (dt_hour_str, job_name, update_time_str))
            return []

        job_path = JobPathHelper(s3baseprefix=self.s3read['s3baseprefix']).get_dirs_by_demand_prediction_name(
            job_name,
            dt_hour_str, 
            update_time_str=s3folder.split('/')[-2])
        
        self.download_folder(job_path['s3folder'], job_path['local_dir'])

        return job_path['local_dir']

    def upload_demand_prediction_result_by_job_name(
        self, 
        job_name,
        dt_hour_str, 
        local_dir):

        update_time_str = self.get_dt_str_by_timestamp(int(datetime.now().timestamp()))['dt']
        job_path = JobPathHelper(s3baseprefix=self.s3write['s3baseprefix']).get_dirs_by_demand_prediction_name(
            job_name,
            dt_hour_str, 
            update_time_str=update_time_str)

        # upload
        self.upload_folder(job_path['s3folder'], local_dir)

    def download_demand_prediction_result(self, dt_hour_str, update_time_str=''):

        return self.download_demand_prediction_result_by_job_name(
            'demand_prediction', 
            dt_hour_str, 
            update_time_str=update_time_str)


    def upload_demand_prediction_result(self, dt_hour_str, local_dir):
        
        return self.upload_demand_prediction_result_by_job_name(
            'demand_prediction',
            dt_hour_str, 
            local_dir)

    #####################################################
    #
    #   etl
    #
    #####################################################

    def get_latest_etl_s3folder(self, s3folders):
        '''
        get latest s3 folder with _SUCCESS file
        '''

        return self.get_latest_s3folder(s3folders)

    #####################################################
    # Vm Status
    #####################################################

    def get_latest_vm_status_s3folder(self, dt_hour_str):
        job_path = ETLPathHelper(s3baseprefix=self.s3read['s3baseprefix']).get_vm_status_dirs(dt_hour_str)
        s3folder_latest = job_path['s3folder_latest']
        if self.is_object_exist('%s%s' % (s3folder_latest, '_SUCCESS')) == True:
            # return latest
            s3folder = s3folder_latest
        else:
            # no latest => search snapshot
            job_prefix = job_path['s3prefix']
            s3folders = self.get_folder_prefixes(job_prefix)
            s3folder = self.get_latest_etl_s3folder(s3folders)

        # return a folder 
        if s3folder == None:
            self.logger.warn('No latest vm_status')
            return None
        else:
            self.logger.info('get latest vm_status s3folder = %s' % s3folder)
            return s3folder

    def get_latest_vm_status(self, dt_hour_str):
        s3folder = self.get_latest_vm_status_s3folder(dt_hour_str)
        if s3folder is None:
            self.logger.warn('No latest_vm_status in %s', dt_hour_str)
            return []

        job_path = ETLPathHelper(s3baseprefix=self.s3read['s3baseprefix']).get_vm_status_dirs(
            dt_hour_str, 
            update_time_str=s3folder.split('/')[-2])

        self.download_folder(job_path['s3folder'], job_path['local_dir'])

        # read local files
        files = os.listdir(job_path['local_dir'])
        data = []
        for file in files:
            if file[-5:] != '.json':
                continue
            path = '%s%s' % (job_path['local_dir'], file)
            with open(path) as fp:
                json_data = json.load(fp)
                #data.extend(json_data['data'])
                data = json_data['data']

        # remove local files
        self.rmdir(job_path['local_dir'])

        if isinstance(data, list)==False:
            data = [data]
        self.logger.info('Read %s logs from vm_status in %s' % (len(data), dt_hour_str))
        self.logger.info('vm_status[0] = %s' % data[0])
        return data

    def write_vm_status(self, item_dict_list, incremental=True):
        if (len(item_dict_list))==0:
            self.logger.warn('Empty input of vm status')
            return

        # data preprocessing
        dict_dt_hour = {}
        for item_dict in item_dict_list:
            dt_hour_str = self.get_dt_str_by_timestamp(item_dict['snap_time'])['dt_hour_str']
            if dt_hour_str in dict_dt_hour:
                dict_dt_hour[dt_hour_str].append(item_dict)
            else:
                dict_dt_hour[dt_hour_str] = [item_dict]

        # upload data to s3
        response = []
        for dt_hour_str, item_list in dict_dt_hour.items():
            etl_key = ETLPathHelper(s3baseprefix=self.s3write['s3baseprefix']).get_etl_key('vm_status')

            if incremental==True:
                # get latest data and join with current item list
                new_list = item_list + self.get_latest_vm_status(dt_hour_str)
                # remove duplicate
                new_list = list({i['_id']:i for i in new_list}.values())        # TODO: should adopt 'snapshot_id' later
            else:
                new_list = item_list

            self.logger.info('write %s vm_status in %s' % (len(new_list), dt_hour_str))
            response.append(self.write_etl_data(etl_key, new_list, dt_hour_str))

        return response

    #####################################################
    # Battery Swap Log
    #####################################################

    def get_latest_battery_swap_log_s3folder(self, dt_hour_str):
        job_path = ETLPathHelper(s3baseprefix=self.s3read['s3baseprefix']).get_battery_swap_log_dirs(dt_hour_str)
        s3folder_latest = job_path['s3folder_latest']
        if self.is_object_exist('%s%s' % (s3folder_latest, '_SUCCESS')) == True:
            # return latest
            s3folder = s3folder_latest
        else:
            # no latest => search snapshot
            job_prefix = job_path['s3prefix']
            s3folders = self.get_folder_prefixes(job_prefix)
            s3folder = self.get_latest_etl_s3folder(s3folders)

        # return a folder 
        if s3folder == None:
            self.logger.warn('No latest battery_swap_log')
            return None
        else:
            self.logger.info('get latest battery_swap_log s3folder = %s' % s3folder)
            return s3folder

    def get_latest_battery_swap_log(self, dt_hour_str):
        s3folder = self.get_latest_battery_swap_log_s3folder(dt_hour_str)
        if s3folder is None:
            self.logger.warn('No latest_battery_swap_log in %s', dt_hour_str)
            return []

        job_path = ETLPathHelper(s3baseprefix=self.s3read['s3baseprefix']).get_battery_swap_log_dirs(
            dt_hour_str, 
            update_time_str=s3folder.split('/')[-2])

        self.download_folder(job_path['s3folder'], job_path['local_dir'])

        # read local files
        files = os.listdir(job_path['local_dir'])
        data = []
        for file in files:
            if file[-5:] != '.json':
                continue
            path = '%s%s' % (job_path['local_dir'], file)
            with open(path) as fp:
                json_data = json.load(fp)
                data = json_data['data']

        # remove local files
        self.rmdir(job_path['local_dir'])

        if isinstance(data, list)==False:
            data = [data]
        self.logger.info('Read %s logs from battery_swap_log in %s' % (len(data), dt_hour_str))
        self.logger.info('battery_swap_log[0] = %s' % data[0])
        return data

    def write_battery_swap_log(self, item_dict_list, incremental=True):
        if (len(item_dict_list))==0:
            self.logger.warn('Empty input of battey swap log')
            return []

        # data preprocessing
        dict_dt_hour = {}
        for item_dict in item_dict_list:
            dt_hour_str = self.get_dt_str_by_timestamp(item_dict['exchange_time'])['dt_hour_str']
            if dt_hour_str in dict_dt_hour:
                dict_dt_hour[dt_hour_str].append(item_dict)
            else:
                dict_dt_hour[dt_hour_str] = [item_dict]
        
        # upload data to s3
        response = []
        for dt_hour_str, item_list in dict_dt_hour.items():
            etl_key = ETLPathHelper(s3baseprefix=self.s3write['s3baseprefix']).get_etl_key('battery_swap_log')

            if incremental==True:
                # get latest data and join with current item list
                new_list_all = item_list + self.get_latest_battery_swap_log(dt_hour_str)
                # remove duplicate
                #new_list = list({i['_id']:i for i in new_list}.values())    # TODO: should adopt 'swap_id' later
                new_list = list({i['swap_id']:i for i in new_list_all}.values())
                if len(new_list_all) != len(new_list):
                    self.logger.warn('found duplicate battery_swap_log in %s: total # = %s, unique #= %s' % (dt_hour_str, len(new_list_all), len(new_list)))
            else:
                new_list = item_list
            
            self.logger.info('write %s battery_swap_log in %s' % (len(new_list), dt_hour_str))
            response.append(self.write_etl_data(etl_key, new_list, dt_hour_str))

        return response

    #####################################################
    # general etl (by Vm-Hour)
    #####################################################

    def get_latest_etl_data_s3folder(self, etl_key, dt_hour_str):
        job_path = ETLPathHelper(s3baseprefix=self.s3read['s3baseprefix']).get_dirs_by_etl_key(etl_key, dt_hour_str)
        s3folder_latest = job_path['s3folder_latest']
        if self.is_object_exist('%s%s' % (s3folder_latest, '_SUCCESS')) == True:
            # return latest
            s3folder = s3folder_latest
        else:
            # no latest => search snapshot
            job_prefix = job_path['s3prefix']
            s3folders = self.get_folder_prefixes(job_prefix)
            s3folder = self.get_latest_etl_s3folder(s3folders)

        # return a folder 
        if s3folder == None:
            self.logger.warn('No latest %s' % etl_key)
            return None
        else:
            self.logger.info('get latest %s s3folder = %s' % (etl_key, s3folder))
            return s3folder

    def get_latest_etl_data(self, etl_key, dt_hour_str):
        s3folder = self.get_latest_etl_data_s3folder(etl_key, dt_hour_str)

        if s3folder is None:
            self.logger.warn('No latest_%s in %s', (etl_key, dt_hour_str))
            return []

        job_path = ETLPathHelper(s3baseprefix=self.s3read['s3baseprefix']).get_dirs_by_etl_key(
            etl_key,
            dt_hour_str, 
            update_time_str=s3folder.split('/')[-2])

        self.download_folder(job_path['s3folder'], job_path['local_dir'])
        
        # read local files
        files = os.listdir(job_path['local_dir'])
        data = None
        for file in files:
            if file[-5:] != '.json':
                continue
            path = '%s%s' % (job_path['local_dir'], file)
            with open(path) as fp:
                json_data = json.load(fp)
                data = json_data['data']

        # remove local files
        self.rmdir(job_path['local_dir'])

        self.logger.info('Read %s logs from %s in %s' % (len(data), etl_key, dt_hour_str))
        return data 
    
    def write_etl_data(self, etl_key, data, dt_hour_str):
        dt_now_str = self.get_dt_str_by_timestamp(int(datetime.now().timestamp()))['dt']
        etl_path = ETLPathHelper(s3baseprefix=self.s3write['s3baseprefix']).get_dirs_by_etl_key(etl_key, dt_hour_str, update_time_str=dt_now_str)
        # write local file and upload
        self.mkdir(etl_path['local_dir'])
        json_data = {}
        json_data['data'] = data
        with open('%s%s.json' % (etl_path['local_dir'], dt_now_str), 'w') as fp:
            json.dump(json_data, fp)
        # upload
        self.upload_folder(etl_path['s3folder'], etl_path['local_dir'])
        # FIXME
        self.delete_folder(etl_path['s3folder_latest'])
        self.upload_folder(etl_path['s3folder_latest'], etl_path['local_dir'])
        # remove local file
        self.rmdir(etl_path['local_dir'])

        return {
            'partition': dt_hour_str,
            's3key': etl_path['s3folder'],
            's3fullpath': 's3://%s/%s' % (self.s3write['s3bucket'], etl_path['s3folder']),
            's3key_latest': etl_path['s3folder_latest'],
            's3fullpath_latest': 's3://%s/%s' % (self.s3write['s3bucket'], etl_path['s3folder_latest']),
        }

    #####################################################
    # specific etl (by Vm-Hour)
    #####################################################

    def get_latest_vm_battery_count_all_s3folder(self, dt_hour_str):
        etl_key = ETLPathHelper(s3baseprefix=self.s3write['s3baseprefix']).get_etl_key('vm_battery_count_all')
        return self.get_latest_etl_data_s3folder(etl_key, dt_hour_str)

    def get_latest_vm_battery_count_s3folder(self, dt_hour_str):
        etl_key = ETLPathHelper(s3baseprefix=self.s3write['s3baseprefix']).get_etl_key('vm_battery_count')
        return self.get_latest_etl_data_s3folder(etl_key, dt_hour_str)

    def get_latest_vm_exchange_count_s3folder(self, dt_hour_str):
        etl_key = ETLPathHelper(s3baseprefix=self.s3write['s3baseprefix']).get_etl_key('vm_exchange_count')
        return self.get_latest_etl_data_s3folder(etl_key, dt_hour_str)
    
    def get_latest_vm_battery_count_all(self, dt_hour_str):
        etl_key = ETLPathHelper(s3baseprefix=self.s3write['s3baseprefix']).get_etl_key('vm_battery_count_all')
        return self.get_latest_etl_data(etl_key, dt_hour_str)

    def get_latest_vm_battery_count(self, dt_hour_str):
        s3folder = self.get_latest_vm_battery_count_s3folder(dt_hour_str)

        if s3folder is None:
            self.logger.warn('No latest_vm_battery_count in %s', dt_hour_str)
            return []

        job_path = ETLPathHelper(s3baseprefix=self.s3read['s3baseprefix']).get_vm_battery_count_dirs(
            dt_hour_str, 
            update_time_str=s3folder.split('/')[-2])

        self.download_folder(job_path['s3folder'], job_path['local_dir'])
        
        # read local files
        files = os.listdir(job_path['local_dir'])
        data = None
        for file in files:
            if file[-5:] != '.json':
                continue
            path = '%s%s' % (job_path['local_dir'], file)
            with open(path) as fp:
                json_data = json.load(fp)
                data = json_data['data']

        # remove local files
        self.rmdir(job_path['local_dir'])

        self.logger.info('Read %s logs from vm_battery_count in %s' % (len(data), dt_hour_str))
        #self.logger.info('vm_battery_count.keys() = %s' % list(data.keys()))
        return data 

    def get_latest_vm_exchange_count(self, dt_hour_str):
        s3folder = self.get_latest_vm_exchange_count_s3folder(dt_hour_str)

        if s3folder is None:
            self.logger.warn('No latest_vm_exchange_count in %s', dt_hour_str)
            return []

        job_path = ETLPathHelper(s3baseprefix=self.s3read['s3baseprefix']).get_vm_exchange_count_dirs(
            dt_hour_str, 
            update_time_str=s3folder.split('/')[-2])

        self.download_folder(job_path['s3folder'], job_path['local_dir'])

        # read local files
        files = os.listdir(job_path['local_dir'])
        data = []
        for file in files:
            if file[-5:] != '.json':
                continue
            path = '%s%s' % (job_path['local_dir'], file)
            with open(path) as fp:
                json_data = json.load(fp)
                data = json_data['data']

        # remove local files
        self.rmdir(job_path['local_dir'])

        self.logger.info('Read %s logs from vm_exchange_count in %s' % (len(data), dt_hour_str))
        self.logger.info('vm_exchange_count[0] = %s' % data[0])
        return data 

    def write_vm_battery_count_all(self, data, dt_hour_str):
        if len(data)==0:
            self.logger.warn('Empty input of vm_battery_count_all')
            return []
        etl_key = ETLPathHelper(s3baseprefix=self.s3write['s3baseprefix']).get_etl_key('vm_battery_count_all')
        self.logger.info('write %s vm_battery_count_all in %s' % (len(data), dt_hour_str))
        return [self.write_etl_data(etl_key, data, dt_hour_str)]

    def write_vm_battery_count(self, data, dt_hour_str):
        if len(data)==0:
            self.logger.warn('Empty input of vm_battery_count')
            return []
        etl_key = ETLPathHelper(s3baseprefix=self.s3write['s3baseprefix']).get_etl_key('vm_battery_count')
        self.logger.info('write %s vm_battery_count in %s' % (len(data), dt_hour_str))
        return [self.write_etl_data(etl_key, data, dt_hour_str)]

    def write_vm_exchange_count(self, data, dt_hour_str):
        if len(data)==0:
            self.logger.warn('Empty input of vm_exchange_count')
            return []
        etl_key = ETLPathHelper(s3baseprefix=self.s3write['s3baseprefix']).get_etl_key('vm_exchange_count')
        self.logger.info('write %s vm_exchange_count in %s' % (len(data), dt_hour_str))
        return [self.write_etl_data(etl_key, data, dt_hour_str)]

    
    #####################################################
    #
    #   general
    #
    #####################################################

    def get_dt_str_by_timestamp(self, timestamp):
        '''
        Return a string format dict by given timestamp
        '''
        dt_str = {}
        dt_str['dt'] = datetime.utcfromtimestamp(timestamp).strftime('%Y-%m-%dT%H:%M:%S')
        dt_str['dt_hour_str'] = dt_str['dt'][:13]
        dt_str['dt_minute_str'] = dt_str['dt'][:16]
        dt_str['time_str'] = dt_str['dt'][-8:]
        return dt_str

    



from mlcore.datahelper.athena.athena_helper import AthenaHelper
from mlcore.datahelper.athena.mllearningdb.create import Creator
from mlcore.datahelper.athena.mllearningdb.update import Updater

from datetime import datetime
import json

import sys
import logging

logging.basicConfig(stream=sys.stdout,
                    level=logging.INFO,
                    format='%(asctime)-s %(levelname)-7s %(name)-s.%(funcName)s(): %(message)s',
                    datefmt='%Y-%m-%dT%H:%M:%S')


class DataHelper(AthenaHelper):

    def __init__(self, config):

        self.config = config['mlcore']['datahelper']['athena']['mllearningdb']
        super().__init__(self.config)
        self.logger = logging.getLogger(self.__class__.__name__)

    def connect(self):
        super().connect()


    def create_table(self, table, database):
        sql = {
            'battery_swap_log': Creator().create_battery_swap_log(database),
            'vm_status': Creator().create_vm_status(database),
            'vm_battery_count': Creator().create_vm_battery_count(database),
            'vm_exchange_count': Creator().create_vm_exchange_count(database),
            'demand_prediction': Creator().create_demand_prediction(database)
        }[table]
        self.query(sql, database)

    def add_partition(self, table, database, pkey, pvalue, s3path):
        sql = Updater().add_partition(table, database, pkey, pvalue, s3path)
        self.query(sql, database)

        sql = Updater().set_partition_location(table, database, pkey, pvalue, s3path)
        self.query(sql, database)

    



from mlcore.datahelper.api.keycloak_sdk import KeycloakSDK
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
import json
import base64
import math
import sys
import logging

logging.basicConfig(stream=sys.stdout,
                    level=logging.INFO,
                    format='%(asctime)-s %(levelname)-7s %(name)-s.%(funcName)s(): %(message)s',
                    datefmt='%Y-%m-%dT%H:%M:%S')




class APIHelper():
    def __init__(self, config):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.config = config['mlcore']['datahelper']['api']['dataplatform']

        self.keycloakSDK = KeycloakSDK(self.config['keycloak'])
        self.keycloak = self.config['keycloak']
        self.access_token = None

        # limit per page
        self.limit_per_page = 200
        self.limit_per_page = {
            'default': 200,
            'nearby_vms': 900
        }

    def get_access_token(self):
        return self.keycloakSDK.get_access_token()

    def logout(self):
        self.keycloakSDK.logout()

    #########################################################
    #
    #   MongoDB API
    #
    #########################################################

    def query_swap_logs(self, cond_dict):
        self.access_token = self.get_access_token()

        data = []
        res = self.query_swap_logs_with_page(cond_dict)

        total_count = res['total_count']
        data = res['data']

        num_page = math.ceil(int(total_count) / float(self.limit_per_page['default']))

        # self.logger.info('total_count = %s' % total_count)
        # self.logger.info('num_page = %s' % num_page)
        # self.logger.info('data = %s' % data)
        # self.logger.info('# of data = %s' % len(data))

        for i in range(1, num_page+1):
            res = self.query_swap_logs_with_page(cond_dict, page=i)
            data = data + res['data']
            # self.logger.info('# of data = %s' % len(data))

        self.logout()

        return data


    def query_swap_logs_with_page(self, cond_dict, page=0):
        url = '%s/swap-logs' % self.config['mongodb-url']

        headers = {
            'Go-Client':self.keycloak['resource'],
            'Authorization':'Bearer {}'.format(self.access_token),
        }

        get_data = {}
        # pagination
        get_data["pagination_criteria"] = {
            "offset": 0 + page * self.limit_per_page['default'],
            "limit": self.limit_per_page['default']
        }
        # essential
        if "exchange_time_start" in cond_dict:
            get_data["exchange_time_start"] = cond_dict["exchange_time_start"]
        if "exchange_time_end" in cond_dict:
            get_data["exchange_time_end"] = cond_dict["exchange_time_end"]
        if "source" in cond_dict:
            get_data["source"] = cond_dict["source"]
        # optinal
        if "create_time_start" in cond_dict:
            get_data["create_time_start"] = cond_dict["create_time_start"]
        if "create_time_end" in cond_dict:
            get_data["create_time_end"] = cond_dict["create_time_end"]
        if "swap_id" in cond_dict:
            get_data["swap_id"] = cond_dict["swap_id"]
        if "vm_guid" in cond_dict:
            get_data["vm_guid"] = cond_dict["vm_guid"]
        if "scooter_guid" in cond_dict:
            get_data["scooter_guid"] = cond_dict["scooter_guid"]

        json_obj = {
            'op_code': 'get',
            'get_data': get_data
        }
        json_option = None

        # print('url=%s' % url)
        # print('header=%s' % headers)
        # print('json=%s' % json_obj)
        # self.logger.info('pagination_criteria = %s' % get_data["pagination_criteria"])
        
        r = requests.post(url, headers=headers, json=json_obj)
        if json_option is None:
            json_option = {}
        #self.logger.info('r.content = %s' % r.content)
        res = json.loads(r.content.decode('utf-8'),**json_option)

        if res['code']==0:
            return res
        else:
            self.logger.warn('Failed to query_swap_logs_with_page: %s', res)
            return None

    def query_vm_status(self, cond_dict):
        self.access_token = self.get_access_token()

        data = []
        res = self.query_vm_status_with_page(cond_dict)

        total_count = res['total_count']
        data = res['data']

        num_page = math.ceil(int(total_count) / float(self.limit_per_page['default']))

        # self.logger.info('total_count = %s' % total_count)
        # self.logger.info('num_page = %s' % num_page)
        # self.logger.info('data = %s' % data)
        # self.logger.info('# of data = %s' % len(data))

        for i in range(1, num_page+1):
            res = self.query_vm_status_with_page(cond_dict, page=i)
            data = data + res['data']
            # self.logger.info('# of data = %s' % len(data))

        self.logout()

        return data

    def query_vm_status_with_page(self, cond_dict, page=0):
        url = '%s/vms/statuses' % self.config['mongodb-url']

        headers = {
            'Go-Client':self.keycloak['resource'],
            'Authorization':'Bearer {}'.format(self.access_token),
        }

        get_data = {}
        # pagination
        get_data["pagination_criteria"] = {
            "offset": 0 + page * self.limit_per_page['default'],
            "limit": self.limit_per_page['default']
        }
        # essential
        if "snap_time_start" in cond_dict:
            get_data["snap_time_start"] = cond_dict["snap_time_start"]
        if "snap_time_end" in cond_dict:
            get_data["snap_time_end"] = cond_dict["snap_time_end"]
        # optinal
        if "state" in cond_dict:
            get_data["state"] = cond_dict["state"]
        if "create_time_start" in cond_dict:
            get_data["create_time_start"] = cond_dict["create_time_start"]
        if "create_time_end" in cond_dict:
            get_data["create_time_end"] = cond_dict["create_time_end"]

        json_obj = {
            'op_code': 'get',
            'get_data': get_data
        }
        json_option = None

        # print('url=%s' % url)
        # print('header=%s' % headers)
        # print('json=%s' % json_obj)
        # self.logger.info('pagination_criteria = %s' % get_data["pagination_criteria"])
        
        r = requests.post(url, headers=headers, json=json_obj)
        if json_option is None:
            json_option = {}
        #self.logger.info('r.content = %s' % r.content)
        res = json.loads(r.content.decode('utf-8'),**json_option)

        if res['code']==0:
            return res
        else:
            self.logger.warn('Failed to query_vm_status_with_page: %s', res)
            return None

    #########################################################
    #
    #   Go Platform API
    #
    #########################################################

    def query_vms(self, vm_ids=None):
        return self.query_vms_v1(vm_ids=vm_ids)


    def query_vms_v1(self, vm_ids=None):
        self.access_token = self.get_access_token()

        data = []
        res = self.query_vms_with_page_v1(vm_ids=vm_ids)

        total_count = res['total_count']
        data = res['data']

        num_page = math.ceil(int(total_count) / float(self.limit_per_page['default']))

        # self.logger.info('total_count = %s' % total_count)
        # self.logger.info('num_page = %s' % num_page)
        # self.logger.info('data = %s' % data)
        # self.logger.info('# of data = %s' % len(data))

        for i in range(1, num_page+1):
            res = self.query_vms_with_page_v1(vm_ids=vm_ids, page=i)
            data = data + res['data']
            # self.logger.info('# of data = %s' % len(data))

        self.logout()

        return data

    def query_vms_with_page_v1(self, vm_ids=None, page=0):
        url = '%s/v1/vms' % self.config['server-url']

        headers = {
            'Go-Client':self.keycloak['resource'],
            'Authorization':'Bearer {}'.format(self.access_token),
        }
        
        if vm_ids==None:
            get_data = {
                'fields_type': 'detail',
                "pagination_criteria": {
                    "offset": 0 + page * self.limit_per_page['default'],
                    "limit": self.limit_per_page['default']
                }
            }
        else:
            get_data = {
                'fields_type': 'detail',
                'vm_ids': vm_ids,
                "pagination_criteria": {
                    "offset": 0 + page * self.limit_per_page['nearby_vms'],
                    "limit": self.limit_per_page['nearby_vms']
                }
            }

        json_obj = {
            'op_code': 'get',
            'get_data': get_data
        }
        json_option = None
        # print('url=%s' % url)
        # print('header=%s' % headers)
        # print('json=%s' % json_obj)
        r = requests.post(url, headers=headers, json=json_obj)
        if json_option is None:
            json_option = {}
        #self.logger.info('r.content = %s' % r.content)
        res = json.loads(r.content.decode('utf-8'),**json_option)

        if res['code']==0:
            return res
        else:
            self.logger.warn('Failed to query_vms_with_page_v1: %s', res)

        

    def query_esrates(self):
        return self.query_esrates_v1()

    def query_esrates_v1(self):
        url = '%s/v1/es-rates' % self.config['server-url']

        self.access_token = self.get_access_token()

        headers = {
            'Go-Client':self.keycloak['resource'],
            'Authorization':'Bearer {}'.format(self.access_token),
        }
        json_obj = {
            'op_code': 'get',
            'get_data': {
                'get_flag': 1
            }
        }
        json_option = None
        # print('url=%s' % url)
        # print('header=%s' % headers)
        # print('json=%s' % json_obj)
        r = requests.post(url, headers=headers, json=json_obj)
        if json_option is None:
            json_option = {}
        res = json.loads(r.content.decode('utf-8'),**json_option)

        self.logout()

        if res['code']==0:
            return res['data']
        else:
            self.logger.warn('Failed to query_esrates_v1: %s', res)

    def query_vms_nearby(self, vms):
        return self.query_vms_nearby_v1(vms)


    def query_vms_nearby_v1(self, vms):
        '''
        format of returned nearby-vms 
        
        [
            {
                'vm_id': 'Arm2QMNZ', 
                'nearby_vms': [
                    {'vm_id': 'Z4mk6rNP', 'distance': 0.35, 'latitude': 25.046688, 'longitude': 121.535175}, 
                    {'vm_id': 'a6mn7WLo', 'distance': 0.35, 'latitude': 25.046688, 'longitude': 121.535175}, 
                    {'vm_id': 'WKNv1YR6', 'distance': 0.64, 'latitude': 25.046219, 'longitude': 121.538603}
                ]
            }
        ]

        '''
        self.access_token = self.get_access_token()

        data = []
        res = self.query_vms_nearby_with_page_v1()

        total_count = res['total_count']
        data = res['data']

        num_page = math.ceil(int(total_count) / float(self.limit_per_page['nearby_vms']))


        for i in range(1, num_page+1):
            res = self.query_vms_nearby_with_page_v1(page=i)
            if len(res['data']) > 0:
                data = data + res['data']
            # self.logger.info('# of data = %s' % len(data))
        
        # dict{vm_id : vm_guid} from vms data
        vd = {v['vm_id']:v['vm_guid'] for v in vms}
        #for k,v in vd.items():
        #    self.logger.info('%s:%s' %(k,v))
        #self.logger.info(vd['wEmDJ3RZ'])

        # map vm_id to vm_guid
        for d in data:
            d['vm_guid'] = vd[ d['vm_id'] ]
            for nv in d['nearby_vms']:
                nv['vm_guid'] = vd[ nv['vm_id'] ]

        key = 'vm_guid'
        col = 'nearby_vms'
        # group by vm_id
        nearby_data = {}
        for d in data:
            vm_id = d[key]
            if vm_id in nearby_data:
                nearby_data[vm_id].extend(d[col])
            else:
                nearby_data[vm_id] = d[col]

        nearby_list = []
        for k,v in nearby_data.items():
            nearby_list.append({key:k, col:v})

        self.logout()

        return nearby_list

    def query_vms_nearby_with_page_v1(self, page=0):
        url = '%s/v1/vms/nearby-vms' % self.config['server-url']

        headers = {
            'Go-Client':self.keycloak['resource'],
            'Authorization':'Bearer {}'.format(self.access_token),
        }
   
        json_obj = {
            'op_code': 'get',
            'get_data': {
                'fields_type': 'detail',
                "pagination_criteria": {
                    "offset": 0 + page * self.limit_per_page['nearby_vms'],
                    "limit": self.limit_per_page['nearby_vms']
                }
            }
        }

        json_option = None
        # print('url=%s' % url)
        # print('header=%s' % headers)
        #print('json=%s' % json_obj)
        r = requests.post(url, headers=headers, json=json_obj)
        #print(r.content)
        if json_option is None:
            json_option = {}

        res = json.loads(r.content.decode('utf-8'),**json_option)

        if res['code']==0:
            return res
        else:
            self.logger.warn('Failed to query_vms_nearby_with_page_v1: %s', res)




import pymongo
from pymongo import MongoClient
import json
from uuid import UUID
from bson import ObjectId

import sys
import logging

logging.basicConfig(stream=sys.stdout,
                    level=logging.INFO,
                    format='%(asctime)-s %(levelname)-7s %(name)-s.%(funcName)s(): %(message)s',
                    datefmt='%Y-%m-%dT%H:%M:%S')

class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, UUID):
            # if the obj is uuid, we simply return the value of uuid
            #return obj.hex
            return str(obj)
        elif isinstance(obj, ObjectId):
            return str(obj)
        return json.JSONEncoder.default(self, obj)

class MongoDBDataHelper():

    def __init__(self, config):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.config = config
        self.connect()

    def connect(self):
        if 'authmech' in self.config:
            self.client = MongoClient(
                host=self.config['server'], 
                username=self.config['username'],
                password=self.config['password'],
                port=int(self.config['port']),
                authSource=self.config['database'],
                authMechanism=self.config['authmech'])
        else:
            self.client = MongoClient(
                host=self.config['server'], 
                port=int(self.config['port']))
                
        self.logger.info('client = %s' % self.client)

        if 'database' in self.config:
            self.db = self.client[self.config['database']]
        else:
            self.db = None

        return self.client, self.db

    def get_collection(self, collection):
        return self.db[collection]

    def find(self, collection, condition):
        self.logger.info('start finding result for ... ')
        self.logger.info('\tdb = %s' % self.db)
        self.logger.info('\tcollection = %s' % collection)
        self.logger.info('\tcondition = %s' % condition)

        cursor = self.db[collection].find(condition)

        result = []
        for c in cursor:
            # make UUID and ObjectID JSON serializable
            result.append(json.loads(json.dumps(c, cls=JSONEncoder)))

        self.logger.info('result.count = %s' % len(result))
        if len(result) > 0:
            self.logger.info('result[0] = %s' % result[0])

        return result




import boto3
import botocore
import random

import sys
import logging

logging.basicConfig(stream=sys.stdout,
                    level=logging.INFO,
                    format='%(asctime)-s %(levelname)-7s %(name)-s.%(funcName)s(): %(message)s',
                    datefmt='%Y-%m-%dT%H:%M:%S')

class AthenaHelper():

    def __init__(self, config):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.config = config
        self.connect()

    def connect(self):
        ###################################################
        '''
        return athena resource
        '''
        def get_athena_resource(config):
            if 'role-arn' in config and config['role-arn']!='':
                if 'aws-access-key-id' in config and 'aws-secret-access-key' in config:
                    sts_client = boto3.client(
                        'sts',
                        aws_access_key_id=config['aws-access-key-id'],
                        aws_secret_access_key=config['aws-secret-access-key'],
                        region_name=config['region'])
                else:
                    sts_client = boto3.client('sts')

                session_name = '%s-%s'%(config['role-session-name'], str(random.randint(1,100)))
                #self.logger.info('session_name = %s' % session_name)
                assumed_role_object = sts_client.assume_role(
				    RoleArn=config['role-arn'],
				    RoleSessionName=session_name)
                credentials = assumed_role_object['Credentials']
            
                athena = boto3.client(
				    'athena',
				    aws_access_key_id=credentials['AccessKeyId'],
				    aws_secret_access_key=credentials['SecretAccessKey'],
				    aws_session_token=credentials['SessionToken'],
                    region_name=config['region'])
            else:
                if 'aws-access-key-id' in config and 'aws-secret-access-key' in config:
                    athena = boto3.client(
                        'athena',
                        aws_access_key_id=config['aws-access-key-id'],
                        aws_secret_access_key=config['aws-secret-access-key'],
                        region_name=config['region'])
                else:
                    athena = boto3.client(
                        'athena',
                        region_name=config['region'])

            return {
                'athena': athena,
                's3output': config['s3output'],
            }
        ###################################################
        
        self.athena = get_athena_resource(self.config)

    def run_query(self, query, database):
        self.connect()
        query_response = self.athena['athena'].start_query_execution(
            QueryString=query,
            QueryExecutionContext={
                'Database': database
            },
            ResultConfiguration={
                'OutputLocation': self.athena['s3output'],
            }
        )
        return query_response

    def create_db(self, database):
        self.connect()
        query_create_db = '''
CREATE DATABASE IF NOT EXISTS %s
        ''' % (database)
        query_response = self.athena['athena'].start_query_execution(
            QueryString=query_create_db,
            ResultConfiguration={
                'OutputLocation': self.athena['s3output']
            }
        )
        self.logger.info('Create database result is shown in %s%s.txt' % (self.athena['s3output'], query_response['QueryExecutionId']))
        return query_response

    def query(self, query, database):
        self.logger.info('Run query %s (database=%s)' % (query, database))
        query_response = self.run_query(query, database)
        self.logger.info('Query result is shown in %s%s.txt' % (self.athena['s3output'], query_response['QueryExecutionId']))
        return query_response
        
    def drop_table(self, table, database):
        query_drop_table = '''
DROP TABLE %s.%s
        ''' % (database, table)
        self.logger.info('Drop table %s (database=%s)' % (table, database))
        query_response = self.run_query(query_drop_table, database)
        self.logger.info('Drop table result is shown in %s%s.txt' % (self.athena['s3output'], query_response['QueryExecutionId']))
        return query_response








from setuptools import setup, find_packages

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
        name='go-ml-core',
        version='0.1.1.dev22',
	  	description='machine learning core library',
        long_description=long_description,
        long_description_content_type='text/markdown',
	  	packages=find_packages(exclude=['tests']),
		url='http://10.0.1.31/ML/ml-storage/ml-etl/',
		author='henrychi',
		author_email='shyuusaku@gmail.com',
		license='Gogoro Inc.',
		keywords='machine learning',
        classifiers = [
            #   3 - Alpha
            #   4 - Beta
            #   5 - Production/Stable
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'Topic :: Software Development :: Build Tools',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
        ],
		install_requires=[
			'boto3>=1.9.57',
			'pymongo>=3.7.2',
            'PyYAML>=3.13',
            'requests>=2.20.1'
		]
)



from mlcore.datahelper.mssql.mssql_data_helper import MSSQLDataHelper
from mlcore.datahelper.mssql.go_server_production import select as Selector

class DataHelper(MSSQLDataHelper):

    def __init__(
        self, 
        logger=None,
        config_file='go_servers.ini',
        host_name='GoServer_Production_ReadOnly'):
        super().__init__( 
            logger=logger,
            config_file=config_file,
            host_name=host_name)

    def connect(self):
        super().connect()

    def get_vm_list(self):
        return super().get_data(Selector.get_vm_list)
        


import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
import base64
import ast
import sys
import logging

logging.basicConfig(stream=sys.stdout,
                    level=logging.INFO,
                    format='%(asctime)-s %(levelname)-7s %(name)-s.%(funcName)s(): %(message)s',
                    datefmt='%Y-%m-%dT%H:%M:%S')
               

def str_to_dict(str):
    str = str.replace('false', 'False')
    str = str.replace('true', 'True')
    return ast.literal_eval(str)

class KeycloakSDK():

    def __init__(self, keycloak):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.keycloak = keycloak
        self.token = None

    def get_access_token(self):
        # 1. service_account_login
        self.service_account_login()
        # 2. get_authorization
        self.get_authorization()
        # 3. refresh_access_token (schedule)
        #self.refresh_access_token()

        self.logger.info('Service account login: %s' % self.token['access_token'])

        return self.token['access_token']

    def service_account_login(self):
        '''
        Service Account Login
        Method : basicAuth(resource, serviceConfig)
        '''

        url = '%s/realms/%s/protocol/openid-connect/token' % (self.keycloak["auth-server-url"], self.keycloak["realm"])
        
        headers = {
            'Content-Type' : 'application/x-www-form-urlencoded',
            'Authorization': 'Basic %s' % base64.b64encode(bytes('%s:%s'% (self.keycloak["resource"], self.keycloak["credentials"]["secret"]),'utf-8')).decode("utf-8")
        }
        payload = {
            'grant_type' : 'client_credentials'
        }
        # print('url=%s' % url)
        # print('header=%s' % headers)
        # print('data=%s' % payload)
        response = requests.post(url, headers=headers, data=payload, verify=False).text
        
        self.token = str_to_dict(response)

    def logout(self):
        '''
        Logout
        '''
        url = '%s/realms/%s/protocol/openid-connect/logout' % (self.keycloak["auth-server-url"], self.keycloak["realm"])
        
        headers = {
            'Content-Type' : 'application/x-www-form-urlencoded',
            'Authorization': 'Basic %s' % base64.b64encode(bytes('%s:%s'% (self.keycloak["resource"], self.keycloak["credentials"]["secret"]),'utf-8')).decode("utf-8")
        }

        payload = {
            'refresh_token' : self.token['refresh_token']
        }

        # print('url=%s' % url)
        # print('header=%s' % headers)
        # print('data=%s' % payload)

        response = requests.post(url, headers=headers, data=payload, verify=False)

        if response.status_code == 204:
            self.logger.info('Logout successfully: %s' % response.text)
            #self.logger.info('Logout successfully.')
            return
        else:
            self.logger.warn('Logout failed: %s' % response.text)


    def get_authorization(self): 
        '''
        GetAuthorization
        Description: Use following information to get authorization. resource 為目標要取得的 application resource
        '''
        
        url = '%s/realms/%s/protocol/openid-connect/token' % (self.keycloak["auth-server-url"], self.keycloak["realm"]) 
        
        headers = {
            'Content-Type' : 'application/x-www-form-urlencoded',
            'Authorization': 'Bearer %s' % self.token['access_token']
        }
        payload = {
            'grant_type' : 'urn:ietf:params:oauth:grant-type:uma-ticket',
            'audience' : self.keycloak["resource"]
        }
        # print('url=%s' % url)
        # print('header=%s' % headers)
        # print('data=%s' % payload)
        response = requests.post(url, headers=headers, data=payload, verify=False).text
        
        self.token = str_to_dict(response)

    def refresh_access_token(self):
        '''
        RefreshGrant
        Method refreshGrant(grant)
        '''

        url = '%s/realms/%s/protocol/openid-connect/token' % (self.auth_server_url, self.realm) 
        
        headers = {
            'Content-Type' : 'application/x-www-form-urlencoded',
            'Authorization': 'Basic %s' % base64.b64encode(bytes('%s:%s'% (self.resource, self.secret),'utf-8')).decode("utf-8")
        }
        payload = {
            'client_id': self.resource,
            'refresh_token' : self.token['refresh_token'],
            'grant_type' : 'refresh_token'
        }
        # print('url=%s' % url)
        # print('header=%s' % headers)
        # print('data=%s' % payload)
        response = requests.post(url, headers=headers, data=payload, verify=False).text
        
        self.token = str_to_dict(response)


import sys
import logging

logging.basicConfig(stream=sys.stdout,
                    level=logging.INFO,
                    format='%(asctime)-s %(levelname)-7s %(name)-s.%(funcName)s(): %(message)s',
                    datefmt='%Y-%m-%dT%H:%M:%S')

class Updater():
    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    def add_partition(self, table, database, pkey, pvalue, s3path):
        sql = '''
ALTER TABLE %s.%s 
ADD PARTITION (`%s`='%s') 
LOCATION '%s';
        ''' % (database, table, pkey, pvalue, s3path)

        return sql

    def set_partition_location(self, table, database, pkey, pvalue, s3path):
        sql = '''
ALTER TABLE %s.%s  
PARTITION (`%s`='%s')
SET LOCATION '%s';
        ''' % (database, table, pkey, pvalue, s3path)
        return sql